from .diff_fn import diff_fn
from .nonneg import nonneg, weighted_nonneg, masked_nonneg
from .norm1 import norm1, weighted_norm1
from .sum_squares import sum_squares, weighted_sum_squares, least_squares, lsqr_options, cg_options
from .group_norm1 import group_norm1, weighted_group_norm1
from .poisson_norm import poisson_norm, weighted_poisson_norm
from .patch_NLM import patch_NLM
from .sum_entries import sum_entries, zero_prox
# from .matlab_external import matlab_external
